"""
Library of common values and functions used by all other libraries in AutoNode.
"""

import getpass
import json
import logging
import os
import pickle

from pyhmy import (
    Typgpy,
    validator,
    exceptions,
)

user = getpass.getuser()
harmony_dir = f"{os.environ['HOME']}/.hmy"
node_dir = f"{os.environ['HOME']}/harmony_node"
node_sh_log_dir = f"{node_dir}/node_sh_logs"
bls_key_dir = f"{harmony_dir}/blskeys"
imported_wallet_pass_file_dir = f"{os.environ['HOME']}/harmony_wallet_pass"
cli_bin_dir = f"{harmony_dir}/bin"
cli_bin_path = f"{cli_bin_dir}/hmy"
saved_validator_path = f"{os.environ['HOME']}/harmony_validator_config.json"
saved_node_config_path = f"{harmony_dir}/.node_config.p"
tui_path = f"{harmony_dir}/tui"

node_script_source = "https://harmony.one/node.sh"
tui_source = "http://pub.harmony.one.s3.amazonaws.com/release/linux-x86_64/mainnet/static/harmony-tui"

default_cli_passphrase = ""
bls_key_len = 96
b32_addr_len = 42
check_interval = 8  # Estimated block time

msg_tag = "[AutoNode]"

# TODO: Keybase integration for identity
_validator_config_default = {
    "validator-addr": None,
    "name": None,
    "website": None,
    "security-contact": None,
    "identity": None,
    "amount": None,
    "min-self-delegation": None,
    "rate": None,
    "max-rate": None,
    "max-change-rate": None,
    "max-total-delegation": None,
    "details": None,
    "gas-price": "1"
}
validator_config = _validator_config_default.copy()

# Invariant: 'public-bls-keys' does not change after node is started.
_node_config_default = {
    "endpoint": "https://api.s0.b.hmny.io/",
    "network": "testnet",
    "clean": False,
    "shard": None,
    "auto-reset": False,
    "auto-active": False,
    "no-validator": False,
    "archival": False,
    "no-download": False,
    "fast-sync": False,
    "expose-rpc": False,
    "public-bls-keys": [],
    "encrypted-wallet-passphrase": b'',
    "_is_recovering": False  # Only used for auto hard-reset
}
node_config = _node_config_default.copy()


def save_validator_config():
    """
    Do not save invalid validator.
    In worst case, new validator information will be re-prompted on re-init of AutoNode.
    """
    if node_config['no-validator']:
        return
    try:
        for key, value in _validator_config_default.items():
            if key not in validator_config.keys():
                raise KeyError(f"Missing key {key} from validator config.")
        if validator_config['validator-addr'] is not None:  # Only validate if validator addr has been saved
            validator.Validator(validator_config['validator-addr']).load(validator_config)
        config_string = json.dumps(validator_config, indent=4)
    except (exceptions.InvalidValidatorError, json.decoder.JSONDecodeError, KeyError, TypeError) as e:
        log(f"{Typgpy.FAIL}Invalid validator information to save.{Typgpy.ENDC}\n"
            f"Error: {e}.\n"
            f"Validator Config: {json.dumps(validator_config, indent=2)}")
        log(f"{Typgpy.WARNING}NOT saving validator config, continuing...{Typgpy.ENDC}")
        return
    save_protected_file(config_string, saved_validator_path, verbose=False)
    # Make validator config file easily writeable
    os.chmod(saved_validator_path, 0o600)


def load_validator_config():
    """
    Load the saved validator config.

    Raises json.decoder.JSONDecodeError, IOError, PermissionError
    """
    with open(saved_validator_path, 'r', encoding='utf8') as f:
        imported_val_config = json.load(f)
        validator_config.update(imported_val_config)


def save_node_config():
    """
    Do not save invalid node config.
    In worst case, node config will be refreshed with valid info on re-init of AutoNode.
    """
    try:
        for key, value in _node_config_default.items():
            if key not in node_config.keys():
                raise KeyError(f"Missing key {key} from node config.")
        node_config_string = pickle.dumps(node_config)
    except (pickle.PickleError, KeyError, TypeError) as e:
        log(f"{Typgpy.FAIL}Invalid node config to save.{Typgpy.ENDC}\n"
            f"Error: {e}.")
        log(f"{Typgpy.WARNING}NOT saving node config, continuing...{Typgpy.ENDC}")
        return
    save_protected_file(node_config_string, saved_node_config_path, verbose=False)


def load_node_config():
    """
    Load the saved node config.

    Raises pickle.PickleError, IOError, PermissionError
    """
    with open(saved_node_config_path, 'rb') as f:
        imported_node_config = pickle.load(f)
        node_config.update(imported_node_config)


def save_protected_file(content, file_path, verbose=True):
    if os.path.isfile(file_path):
        if os.access(file_path, os.R_OK):  # check for min needed perms.
            os.remove(file_path)
        else:
            raise PermissionError(f"Cannot save protected file to {file_path} for user {user}")
    if isinstance(content, bytes):
        with open(file_path, 'wb') as f:
            f.write(content)
            protect_file(file_path, verbose=verbose)
    else:
        with open(file_path, 'w', encoding='utf8') as f:
            f.write(content)
            protect_file(file_path, verbose=verbose)


def reset_validator_config():
    validator_config.clear()
    validator_config.update(_validator_config_default)


def reset_node_config():
    node_config.clear()
    node_config.update(_node_config_default)


def log(*args):
    """
    Tagged print and log for AutoNode library.
    # TODO: implement and correct msgs for log levels
    """
    logging.getLogger('AutoNode').debug(*args)


def protect_file(file_path, verbose=True):
    """
    Protect a file with chmod 400.
    """
    if verbose:
        log(f"{Typgpy.WARNING}Protecting file `{file_path}` for user {user}{Typgpy.ENDC}")
    os.chmod(file_path, 0o400)



"""
Library for all things related to running a Harmony node with AutoNode.
"""

import glob
import json
import logging
import os
import shutil
import stat
import subprocess
import time

import pyhmy.rpc.exceptions as rpc_exception
import requests
from pyhmy import (
    blockchain,
    cli,
    Typgpy
)

from .common import (
    log,
    node_script_source,
    node_sh_log_dir,
    node_config,
    node_dir,
    bls_key_dir,
    harmony_dir
)
from .util import (
    input_with_print,
    get_simple_rotating_log_handler,
    is_bls_file,
)

node_sh_out_path = f"{node_sh_log_dir}/out.log"
node_sh_err_path = f"{node_sh_log_dir}/err.log"

log_path = f"{harmony_dir}/autonode_node.log"

rclone_space_buffer = 5 * 2 ** 30  # 5GB in bytes
rclone_config = "harmony"


def _node_clean(verbose=True):
    log_dir = f"{node_dir}/latest"
    backup_log_dir = f"{node_dir}/backup_logs/{time.strftime('%Y/%m/%d/%H')}"
    os.makedirs(backup_log_dir, exist_ok=True)
    try:
        if os.path.isdir(log_dir):
            if verbose:
                log(f"{Typgpy.WARNING}[!] Moving {log_dir} to {backup_log_dir} {Typgpy.ENDC}")
            if os.path.isdir(backup_log_dir):
                shutil.rmtree(backup_log_dir)  # Should never happen, always keep latest.
            shutil.move(log_dir, backup_log_dir)
        for directory in filter(lambda e: os.path.isdir(e), os.listdir(node_dir)):
            if directory.startswith("harmony_db_") or directory.startswith(".dht"):
                path = os.path.join(node_dir, directory)
                if verbose:
                    log(f"{Typgpy.WARNING}[!] Removing {path} {Typgpy.ENDC}")
                shutil.rmtree(path)
    except shutil.Error as e:
        raise SystemExit(e)


def _rclone(rclone_sync_dir, shard, verbose=True):
    db_dir = f"{node_dir}/harmony_db_{shard}"
    node_sh_rclone_err_path = f"{node_sh_log_dir}/rclone_err_{shard}.log"
    node_sh_rclone_out_path = f"{node_sh_log_dir}/rclone_out_{shard}.log"

    try:
        # Assumption made on installed rclone config.
        rclone_path = f'harmony://pub.harmony.one/{rclone_sync_dir}/harmony_db_{shard}'
        if verbose:
            log(f"{Typgpy.WARNING}rclone harmony_db_{shard} from {rclone_path} in progress...{Typgpy.ENDC}")
        with open(node_sh_rclone_out_path, 'w') as fo:
            with open(node_sh_rclone_err_path, 'w') as fe:
                return subprocess.Popen(['rclone', 'sync', '-P', rclone_path, db_dir],
                                        env=os.environ, stdout=fo, stderr=fe)
    except OSError as e:
        if verbose:
            log(f"{Typgpy.FAIL}Failed to rclone shard {shard} db, error {e}{Typgpy.ENDC}")
            log(f"{Typgpy.WARNING}Removing shard {shard} db if it exists{Typgpy.ENDC}")
            if os.path.isdir(db_dir):
                shutil.rmtree(db_dir)


def _rclone_space_required(rclone_sync_dir, shard):
    rclone_path = f'{rclone_config}://pub.harmony.one/{rclone_sync_dir}/harmony_db_{shard}'
    try:
        space_output = subprocess.check_output(['rclone', 'size', rclone_path, '--json'], env=os.environ)
        return int(json.loads(space_output.decode('utf8'))['bytes'])
    except (subprocess.CalledProcessError, json.decoder.JSONDecodeError, KeyError) as e:
        log(f"{Typgpy.WARNING}Failed to get rclone db size requirement, error {e}{Typgpy.ENDC}")
        return 0


def _rclone_db(shard, verbose=True):
    # WARNING: rclone sync directory naming convention may change in the future
    if node_config['archival']:
        rclone_sync_dir = node_config['network'] + '.archival'
    else:
        rclone_sync_dir = node_config['network'] + '.min'

    _, _, free_space = shutil.disk_usage(os.environ['HOME'])
    free_space = free_space - rclone_space_buffer

    rclone_processes = []
    if shard == 0:
        required_space = _rclone_space_required(rclone_sync_dir, shard)
        if required_space == 0:
            log(f"{Typgpy.WARNING}Fast-sync db not available.\n"
                f"Skipping rclone shard {shard}...{Typgpy.ENDC}")
            return
        if required_space > free_space:
            log(f"{Typgpy.WARNING}[!] Insufficient disk space. Required: {required_space}, Free: {free_space} \n"
                f"Skipping fast sync...{Typgpy.ENDC}")
            log(f"{Typgpy.WARNING}[!] Suggest increasing disk space before running node!{Typgpy.ENDC}")
            return
        rclone_processes.append(_rclone(rclone_sync_dir, shard, verbose=verbose))
    else:
        required_beacon_space = _rclone_space_required(rclone_sync_dir, 0)
        required_shard_space = _rclone_space_required(rclone_sync_dir, shard)
        total_required_space = required_beacon_space + required_shard_space
        if total_required_space > free_space:
            log(f"{Typgpy.WARNING}[!] Insufficient disk space. Required: {total_required_space}, Free: {free_space} \n"
                f"Skipping fast sync...{Typgpy.ENDC}")
            log(f"{Typgpy.WARNING}[!] Suggest increasing disk space before running node!{Typgpy.ENDC}")
            return
        if required_beacon_space == 0:
            log(f"{Typgpy.WARNING}Fast-sync db not available.\n"
                f"Skipping rclone shard 0...{Typgpy.ENDC}")
        else:
            rclone_processes.append(_rclone(rclone_sync_dir, 0, verbose=verbose))
        if required_shard_space == 0:
            log(f"{Typgpy.WARNING}Fast-sync db not available.\n"
                f"Skipping rclone shard {shard}...{Typgpy.ENDC}")
        else:
            rclone_processes.append(_rclone(rclone_sync_dir, shard, verbose=verbose))

    for p in rclone_processes:
        p.wait()

    if all(p.returncode == 0 for p in rclone_processes):
        if verbose:
            log(f"{Typgpy.OKGREEN}rclone done!{Typgpy.ENDC}")
    else:
        failed_db_list = glob.glob(os.path.join(node_dir, 'harmony_db_*'))
        for db in failed_db_list:
            db_shard = os.path.basename(db).split('_')[-1]
            log(f'{Typgpy.FAIL}[!] Failed to rclone shard {db_shard} db.{Typgpy.ENDC}')
            log(f'{Typgpy.WARNING}Check {os.path.join(node_sh_log_dir, f"rclone_err_{db_shard}.log")}\n'
                f'Removing {db}.{Typgpy.ENDC}')
            shutil.rmtree(db)
        raise SystemExit('Fast sync failed.')


def _get_node_shard():
    """
    Returns node shard based on config.
    Returns None if 0 or > 1 shards are derived from the node's BLS keys.
    """
    key_shards = []
    if not node_config['public-bls-keys']:
        log(f"{Typgpy.WARNING}No saved BLS keys for node!{Typgpy.ENDC}")
    for bls_key in node_config['public-bls-keys']:
        try:
            key_shards.append(json.loads(cli.single_call(['hmy', 'utility', 'shard-for-bls', bls_key,
                                                          '--node', f'{node_config["endpoint"]}']))['shard-id'])
        except json.decoder.JSONDecodeError:
            log(f'{Typgpy.WARNING}[!] Failed to get shard for bls key {bls_key}!{Typgpy.ENDC}')
    if not key_shards:
        return None
    assert len(
        set(key_shards)) == 1, f"Node BLS keys can only be for 1 shard. BLS keys: {node_config['public-bls-keys']}"
    return key_shards[0]


def start(auto=False, verbose=True):
    """
    Start the harmony process and return the PID.

    Note that process is running after function return.
    """
    if subprocess.call(["pgrep", "harmony"], env=os.environ) == 0:
        raise RuntimeError("Harmony process is already running, can only start 1 node on machine with AutoNode.")
    old_logging_handlers = logging.getLogger('AutoNode').handlers.copy()
    logging.getLogger('AutoNode').addHandler(get_simple_rotating_log_handler(log_path))
    log(f"{Typgpy.HEADER}Starting node...{Typgpy.ENDC}")
    if not auto:
        log(f"{Typgpy.WARNING}You are starting a Harmony Node manually, "
            f"this is not recommended, continue? [Y]/n{Typgpy.ENDC}")
        if input_with_print("> ") not in {'Y', 'y', 'yes', 'Yes'}:
            raise SystemExit()
    assert_valid_bls_key_directory()
    os.chdir(node_dir)
    if os.path.isfile(f"{node_dir}/node.sh"):
        os.remove(f"{node_dir}/node.sh")
    try:
        r = requests.get(node_script_source, timeout=30, verify=True)
        with open("node.sh", 'w') as f:
            node_sh = r.content.decode()
            # WARNING: Hack until node.sh is changed for auto-node.
            node_sh = node_sh.replace("save_pass_file=false", 'save_pass_file=true')
            f.write(node_sh)
    except requests.exceptions.RequestException as e:
        raise SystemExit(e)
    st = os.stat("node.sh")
    os.chmod("node.sh", st.st_mode | stat.S_IEXEC)
    node_args = ["./node.sh", "-N", node_config["network"], "-z", "-f", bls_key_dir, "-M", "-S"]
    if node_config['expose-rpc']:
        if verbose:
            log(f"{Typgpy.WARNING}[!] Starting node with exposed RPC ports.{Typgpy.ENDC}")
        node_args.append("-P")
    if node_config['clean']:
        if verbose:
            log(f"{Typgpy.WARNING}[!] Cleaning up old files before starting node.{Typgpy.ENDC}")
        _node_clean(verbose=verbose)
    if node_config['fast-sync']:
        if verbose:
            log(f'{Typgpy.WARNING}[!] Fast syncing before starting node.{Typgpy.ENDC}')
        shard = _get_node_shard()
        if shard is not None:
            _rclone_db(shard, verbose=verbose)
        else:
            log(f'{Typgpy.WARNING}[!] Unable to determine node shard.\n'
                f'Skipping fast sync...{Typgpy.ENDC}')
    if node_config['archival']:
        if verbose:
            log(f"{Typgpy.WARNING}[!] Starting node in archival mode.{Typgpy.ENDC}")
        node_args.append("-A")
    if node_config['no-download']:
        harmony_binary = os.path.realpath(os.path.join(node_dir, 'harmony'))
        if os.path.exists(harmony_binary):
            if 'static' in subprocess.check_output(['file', harmony_binary], env=os.environ).decode('utf8'):
                if verbose:
                    log(f"{Typgpy.WARNING}[!] Starting node with existing harmony binary.{Typgpy.ENDC}")
                node_args.append("-D")
            else:
                log(f"{Typgpy.WARNING}[!] Provided harmony binary is not statically linked. "
                    f"Path: {harmony_binary}{Typgpy.ENDC}")
                raise SystemExit(f"Harmony binary not statically linked. Path: {harmony_binary}")
        else:
            if verbose:
                log(f"{Typgpy.WARNING}[!] Starting node with released harmony binary.{Typgpy.ENDC}")
    with open(node_sh_out_path, 'w') as fo:
        with open(node_sh_err_path, 'w') as fe:
            if verbose:
                log(f"{Typgpy.HEADER}Starting node!{Typgpy.ENDC}")
            logging.getLogger('AutoNode').handlers = old_logging_handlers  # Reset logger to old handlers
            return subprocess.Popen(node_args, env=os.environ, stdout=fo, stderr=fe).pid


# TODO (low prio): create stream load printer for multiple waits_for_node_response
def wait_for_node_response(endpoint, verbose=True, tries=float("inf"), sleep=0.5):
    count = 0
    while True:
        count += 1
        try:
            blockchain.get_latest_header(endpoint=endpoint)
            break
        except (rpc_exception.RequestsError, rpc_exception.RequestsTimeoutError,
                rpc_exception.RPCError):
            if count > tries:
                raise TimeoutError(f"{endpoint} did not respond in {count} attempts (~{sleep * count} seconds)")
            if verbose and count % 10 == 0:
                log(f"{Typgpy.WARNING}Waiting for {endpoint} to respond, tried {count} times "
                    f"(~{sleep * count} seconds waited so far){Typgpy.ENDC}")
            time.sleep(sleep)
    if verbose:
        log(f"{Typgpy.HEADER}[!] {endpoint} is alive!{Typgpy.ENDC}")


def assert_no_invalid_blocks():
    if os.path.isdir(f"{node_dir}/latest"):
        files = glob.glob(f"{node_dir}/latest/zero*.log")
        if files:
            log_path = files[-1]
            assert not has_invalid_block(
                log_path), f"`invalid merkle root` present in {log_path}, restart AutoNode with clean option"


def is_signing(count=1500):
    """
    Read the last `count` lines and check for signing logs.
    """
    if os.path.isdir(f"{node_dir}/latest"):
        files = glob.glob(f"{node_dir}/latest/zero*.log")
        if files:
            log_path = files[-1]
            content = subprocess.check_output(["tail", "-n", str(count), str(log_path)], env=os.environ).decode().split(
                "\n")
            for line in content:
                line = line.rstrip()
                if "BINGO" in line or "HOORAY" in line:
                    return True
    return False


def has_invalid_block(log_file_path):
    assert os.path.isfile(log_file_path), f"{log_file_path} is not a file"
    try:
        with open(log_file_path, 'r', encoding='utf8') as f:
            for line in f:
                line = line.rstrip()
                if "invalid merkle root" in line:
                    return True
    except (UnicodeDecodeError, IOError):
        log(f"{Typgpy.WARNING}WARNING: failed to read `{log_file_path}` to check for invalid block{Typgpy.ENDC}")
    return False


def assert_valid_bls_key_directory():
    """
    Asserts that the BLS keys directory contains the BLS keys for the
    given node config.

    Note that this must match EXACTLY.

    Will raise assertion error if one BLS key is missing OR
    if there is 1 BLS key more than what is in the node config.
    """
    bls_keys = list(filter(lambda e: is_bls_file(e, '.key'), os.listdir(bls_key_dir)))
    bls_pass = list(filter(lambda e: is_bls_file(e, '.pass'), os.listdir(bls_key_dir)))

    assert len(bls_keys) == len(bls_pass), f"number of BLS keys and BLS pass files must be equal in {bls_key_dir}."
    bls_pub_keys = set(x.replace('.key', '') for x in bls_keys)
    assert len(bls_pub_keys) == len(bls_keys), f"sanity check: cannot contain duplicate BLS keys in {bls_key_dir}"
    found_keys_count = len(bls_pub_keys)

    for key in bls_pass:
        bls_key = key.replace('.pass', '')
        assert bls_key in bls_pub_keys, f"got BLS pass for {bls_key}, but BLS key not found in {bls_key_dir}"

    conf_keys_count = len(node_config['public-bls-keys'])
    if conf_keys_count != found_keys_count:
        raise AssertionError(f"number of configured BLS keys ({conf_keys_count}) not equal to number of BLS keys "
                             f"keys ({found_keys_count}) in {bls_key_dir}")
    for key in node_config['public-bls-keys']:
        assert key in bls_pub_keys, f"configured BLS key {key} not found in {bls_key_dir}"


def assert_started(timeout=60, do_log=False):
    """
    Assert the node has started within the given timeout.
    Node that rclone does NOT count towards timeout.
    """
    has_informed_rclone = False
    start_time = time.time()
    while time.time() - start_time < timeout:
        if subprocess.call(["pgrep", "rclone"], env=os.environ, stdout=subprocess.DEVNULL) == 0:
            timeout += 1
            if not has_informed_rclone and do_log:
                log(f"{Typgpy.WARNING}Fast-sync (rclone) is in progress...{Typgpy.ENDC}")
                has_informed_rclone = True
        elif subprocess.call(["pgrep", "harmony"], env=os.environ, stdout=subprocess.DEVNULL) == 0:
            if do_log:
                log(f"{Typgpy.OKGREEN}Harmony node is running...{Typgpy.ENDC}")
            return
        time.sleep(1)
    if do_log:
        log(f"{Typgpy.FAIL}Harmony node is NOT running!{Typgpy.ENDC}")
    raise AssertionError("Node failed to start")



"""
Library of custom exceptions raised by AutNode
"""

from cryptography.fernet import InvalidToken
from pyhmy import Typgpy

from .common import (
    node_config,
    save_node_config
)


class ResetNode(Exception):
    """
    The only exception that triggers a hard reset.
    """

    def __init__(self, *args, clean=False):
        node_config['clean'] = clean
        save_node_config()
        super(ResetNode, self).__init__(*args)


class InvalidWalletPassphrase(InvalidToken):
    """
    Exception raised if passphrase is invalid
    """

    def __init__(self):
        super(InvalidToken, self).__init__(f"Re-authenticate wallet passphrase with "
                                           f"'{Typgpy.OKGREEN}auto-node auth-wallet{Typgpy.ENDC}'")



"""
Library of common utils used by most function libraries of AutoNode.

To prevent cyclic import minimize the importing of other libraries in AutoNode.
"""

import gzip
import signal
import getpass
import logging
import logging.handlers
import os
import sys

import pexpect
from pyhmy import (
    Typgpy,
    json_load
)
from pyhmy import cli

from .common import (
    log,
    node_config,
    validator_config,
    msg_tag,
    bls_key_len
)
from .exceptions import (
    InvalidWalletPassphrase
)
from .passphrase import (
    decrypt_wallet_passphrase,
    is_valid_passphrase
)


class Timeout:
    def __init__(self, seconds=1, error_message='Timeout'):
        self.seconds = seconds
        self.error_message = error_message

    def handle_timeout(self, signum, frame):
        raise TimeoutError(self.error_message)

    def __enter__(self):
        signal.signal(signal.SIGALRM, self.handle_timeout)
        signal.alarm(self.seconds)

    def __exit__(self, type, value, traceback):
        signal.alarm(0)


class _GZipRotator:
    """A simple zip rotator for logging"""

    def __call__(self, source, dest):
        os.rename(source, dest)
        f_in = open(dest, 'rb')
        f_out = gzip.open("%s.gz" % dest, 'wb')
        f_out.writelines(f_in)
        f_out.close()
        f_in.close()
        os.remove(dest)


def get_wallet_passphrase():
    """
    Gets encrypted passphrase from node_config, unencrypt, validate, and return.

    Raises InvalidWalletPassphrase if encrypted passphrase is invalid.
    """
    try:
        passphrase = decrypt_wallet_passphrase(node_config["encrypted-wallet-passphrase"])
    except InvalidWalletPassphrase:
        with Timeout(seconds=60*2, error_message="Timeout for user input"):
            passphrase = getpass.getpass(f"Enter wallet passphrase for {validator_config['validator-addr']}\n> ")
    if not is_valid_passphrase(passphrase, validator_config["validator-addr"]):
        raise InvalidWalletPassphrase()
    return passphrase


def pexpect_input_wallet_passphrase(proc, passphrase, prompt="Enter wallet keystore passphrase:\r\n"):
    """
    Interactively input the wallet passphrase to the given pexpect child process
    """
    assert isinstance(proc, pexpect.pty_spawn.spawn)
    proc.expect(prompt)
    proc.sendline(passphrase)


def pexpect_input_wallet_creation_passphrase(proc, passphrase):
    """
    This will enter the `passphrase` interactively given the pexpect child program, `proc`.
    """
    assert isinstance(proc, pexpect.pty_spawn.spawn)
    pexpect_input_wallet_passphrase(proc, passphrase, prompt="Enter passphrase:\r\n")
    pexpect_input_wallet_passphrase(proc, passphrase, prompt="Repeat the passphrase:\r\n")
    proc.expect("\n")


def input_with_print(prompt_str, auto_interaction=None):
    if auto_interaction is not None:
        return auto_interaction
    user_input = input(prompt_str).strip()
    sys.stdout.write(f"{Typgpy.ENDC}")
    sys.stdout.flush()
    log(f"{Typgpy.OKBLUE}You entered: `{Typgpy.OKGREEN}{user_input}{Typgpy.OKBLUE}`{Typgpy.ENDC}")
    return user_input


def check_min_bal_on_s0(address, amount, endpoint=node_config['endpoint'], timeout=30):
    balances = json_load(cli.single_call(['hmy', '--node', endpoint, 'balances', address], timeout=timeout))
    for bal in balances:
        if bal['shard'] == 0:
            return bal['amount'] >= amount


def get_simple_rotating_log_handler(log_file_path, max_size=5 * 1024 * 1024):
    """
    A simple log handler with no level support.
    Used purely for the output rotation.

    `max_size` of log file is in bytes.
    """
    log_formatter = logging.Formatter(f'{msg_tag} %(message)s')
    handler = logging.handlers.RotatingFileHandler(log_file_path, mode='a', maxBytes=max_size,
                                                   backupCount=5, encoding=None, delay=0)
    handler.setLevel(logging.DEBUG)
    handler.setFormatter(log_formatter)
    handler.rotator = _GZipRotator()
    return handler


def shard_for_bls(public_bls_key):
    """
    Fetch the shard for the BLS key.

    NOTE: Expect to be invalid one we have resharding.
    """
    return json_load(cli.single_call(['hmy', '--node', f'{node_config["endpoint"]}', 'utility',
                                      'shard-for-bls', public_bls_key]))['shard-id']


def is_bls_file(file_name, suffix):
    if file_name.startswith('.') or not file_name.endswith(suffix):
        return False
    tok = file_name.split(".")
    if len(tok) != 2 or len(tok[0]) != bls_key_len:
        return False
    return True



"""
Library to take care of all passphrase encryption used by AutoNode.
"""

import base64
import os
import subprocess

import pexpect
from cryptography.fernet import Fernet
from cryptography.fernet import InvalidToken
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, hmac
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from pyhmy import (
    cli,
    Typgpy
)

from .common import (
    validator_config,
    node_config,
    log
)
from .exceptions import (
    InvalidWalletPassphrase
)


def _get_harmony_pid():
    try:
        return subprocess.check_output(["pgrep", "harmony"], env=os.environ).strip()
    except (subprocess.TimeoutExpired, subprocess.CalledProcessError):
        return b'0'


def _get_process_info(pid):
    assert isinstance(pid, bytes)
    pid = pid.strip()
    try:
        if int(pid) <= 0:
            return b'0'
        proc_start = subprocess.check_output(["ps", "-p", pid.decode(), "-o", "lstart="], env=os.environ).strip()
        proc_command = subprocess.check_output(["ps", "-p", pid.decode(), "-o", "command="], env=os.environ).strip()
        return proc_start + b' ' + proc_command
    except (subprocess.TimeoutExpired, subprocess.CalledProcessError, ValueError):
        return b'0'


def _get_node_based_salt():
    key = ''.join([str(x) for x in node_config["public-bls-keys"]]
                  + [str(validator_config["validator-addr"]), str(validator_config["identity"])]).encode()
    return hmac.HMAC(key, hashes.SHA256(), backend=default_backend()).finalize()


def _derive_wallet_encryption_key():
    """
    Create the wallet encryption key based on:
        PBKDF2HMAC(PID(harmony) + ProcessInfo(harmony), salt=HMAC(node_bls_public_keys, Validator_Addr, Validator_ID))

    This means that the encryption key is only valid for when AutoNode has a harmony node process running.
    """
    pid = _get_harmony_pid().strip()
    proc_info = _get_process_info(pid).strip()
    salt = _get_node_based_salt().strip()
    kdf = PBKDF2HMAC(
        algorithm=hashes.SHA256(),
        length=32,
        salt=salt,
        iterations=10000,
        backend=default_backend()
    )
    data = pid + b' ' + proc_info
    return base64.urlsafe_b64encode(kdf.derive(data))


def encrypt_wallet_passphrase(passphrase):
    """
    Encrypt the given passphrase.

    Returned string can be stored on disk, will be invalidated once harmony process is stopped.
    """
    assert isinstance(passphrase, str)
    return Fernet(_derive_wallet_encryption_key()).encrypt(passphrase.encode())


def decrypt_wallet_passphrase(encrypted_wallet_passphrase):
    """
    Decrypt the given encrypted passphrase.
    """
    assert isinstance(encrypted_wallet_passphrase, bytes)
    try:
        return Fernet(_derive_wallet_encryption_key()).decrypt(encrypted_wallet_passphrase).decode()
    except InvalidToken:
        raise InvalidWalletPassphrase()


def is_valid_passphrase(passphrase, validator_address):
    """
    Validate the given passphrase, can be an expensive call.
    """
    cmd = ["hmy", "keys", "check-passphrase", validator_address]
    try:
        proc = cli.expect_call(cmd)
        proc.expect("Enter wallet keystore passphrase:\r\n")
        proc.sendline(passphrase)
        proc.expect("Valid passphrase\r\n")
        proc.expect(pexpect.EOF)
        return True
    except RuntimeError as e:
        log(f"{Typgpy.FAIL}Failed to verify passphrase due to: {e}{Typgpy.ENDC}")
        return False
    except pexpect.ExceptionPexpect as e:
        log(f"{Typgpy.FAIL}Failed to verify passphrase due to:\n{e.get_trace()}{Typgpy.ENDC}")
        return False



"""
This library takes care of all validator related commands.
"""

import json
import logging
import subprocess
import sys
import time
import traceback
from decimal import Decimal

import pexpect
from pyhmy import (
    blockchain,
    cli,
    staking,
    account,
    numbers,
    Typgpy,
    exceptions
)

from .common import (
    log,
    bls_key_dir,
    node_config,
    validator_config,
    check_interval,
)
from .initialize import (
    setup_validator_config,
    setup_wallet_passphrase,
)
from .node import (
    log_path,
    wait_for_node_response,
    assert_no_invalid_blocks,
    assert_started as assert_node_started,
    is_signing
)
from .util import (
    check_min_bal_on_s0,
    input_with_print,
    get_simple_rotating_log_handler,
    get_wallet_passphrase,
    pexpect_input_wallet_passphrase
)

_balance_buffer = Decimal(1)
_hard_reset_recovery = False


def _interaction_preprocessor(hard_reset_recovery):
    """
    All user calls (i.e: validator setup) must be processed by this
    """
    global _hard_reset_recovery
    _hard_reset_recovery = hard_reset_recovery
    old_logging_handlers = logging.getLogger('AutoNode').handlers.copy()
    logging.getLogger('AutoNode').addHandler(get_simple_rotating_log_handler(log_path))
    if node_config['no-validator']:
        print(f"{Typgpy.WARNING}Config specifies no validator automation, exiting...{Typgpy.ENDC}")
        exit(0)
    return old_logging_handlers


def _add_bls_key_to_validator():
    """
    Assumes past staking epoch by definition of adding keys to existing validator
    """
    _verify_account_balance(0.1 * len(node_config["public-bls-keys"]))  # Heuristic amount for balance
    chain_val_info = get_validator_information()
    bls_keys = set(x.replace('0x', '') for x in chain_val_info["validator"]["bls-public-keys"])
    for k in (x.replace('0x', '') for x in node_config["public-bls-keys"]):
        if k not in bls_keys:  # Add imported BLS key to existing validator if needed
            _send_edit_validator_tx(k)
        else:
            log(f"{Typgpy.WARNING}Bls key: {Typgpy.OKGREEN}{k}{Typgpy.WARNING} "
                f"is already present, ignoring...{Typgpy.ENDC}")


def _send_edit_validator_tx(bls_key_to_add):
    log(f"{Typgpy.OKBLUE}Adding bls key: {Typgpy.OKGREEN}{bls_key_to_add}{Typgpy.OKBLUE} "
        f"to validator: {Typgpy.OKGREEN}{validator_config['validator-addr']}{Typgpy.ENDC}")
    passphrase = get_wallet_passphrase()
    count = 0
    while True:
        count += 1
        try:
            cmd = ['hmy', '--node', f'{node_config["endpoint"]}', 'staking', 'edit-validator',
                   '--validator-addr', f'{validator_config["validator-addr"]}',
                   '--add-bls-key', bls_key_to_add, '--bls-pubkeys-dir', bls_key_dir, '--passphrase']
            if validator_config["gas-price"]:
                cmd.extend(['--gas-price', f'{validator_config["gas-price"]}'])
            proc = cli.expect_call(cmd)
            pexpect_input_wallet_passphrase(proc, passphrase)
            proc.expect(pexpect.EOF)
            response = proc.before.decode()
            log(f"{Typgpy.OKBLUE}Edit-validator transaction response: "
                f"{Typgpy.OKGREEN}{response}{Typgpy.ENDC}")
            return
        except (RuntimeError, TimeoutError, ConnectionError, subprocess.CalledProcessError) as e:
            log(f"{Typgpy.FAIL}Edit-validator transaction failure (attempt {count}). Error: {e}{Typgpy.ENDC}")
            if not _hard_reset_recovery:
                raise e
            log(f"{Typgpy.WARNING}Trying again in {check_interval} seconds.{Typgpy.ENDC}")
            time.sleep(check_interval)


def _verify_staking_epoch():
    """
    Invariant: All staking transactions are done AFTER staking epoch.
    """
    log(f"{Typgpy.OKBLUE}Verifying Staking Epoch...{Typgpy.ENDC}")
    staking_epoch = blockchain.get_staking_epoch(endpoint=node_config['endpoint'])
    curr_epoch = blockchain.get_current_epoch(endpoint=node_config['endpoint'])
    while curr_epoch < staking_epoch:  # WARNING: using staking epoch for extra security of configs.
        sys.stdout.write(f"\rWaiting for staking epoch ({staking_epoch}) -- current epoch: {curr_epoch}")
        sys.stdout.flush()
        time.sleep(check_interval)
        curr_epoch = blockchain.get_current_epoch(endpoint=node_config['endpoint'])
    log(f"{Typgpy.OKGREEN}Network is at or past staking epoch{Typgpy.ENDC}")


def _verify_prestaking_epoch():
    """
    Invariant: All staking transactions are done AFTER staking epoch.
    """
    log(f"{Typgpy.OKBLUE}Verifying Pre Staking Epoch...{Typgpy.ENDC}")
    prestaking_epoch = blockchain.get_prestaking_epoch(endpoint=node_config['endpoint'])
    curr_epoch = blockchain.get_current_epoch(endpoint=node_config['endpoint'])
    while curr_epoch < prestaking_epoch:
        sys.stdout.write(f"\rWaiting for pre staking epoch ({prestaking_epoch}) -- current epoch: {curr_epoch}")
        sys.stdout.flush()
        time.sleep(check_interval)
        curr_epoch = blockchain.get_current_epoch(endpoint=node_config['endpoint'])
    log(f"{Typgpy.OKGREEN}Network is at or past pre staking epoch{Typgpy.ENDC}")


def _verify_account_balance(amount):
    count = 0
    log(f"{Typgpy.OKBLUE}Verifying Balance...{Typgpy.ENDC}")
    while True:
        count += 1
        if not check_min_bal_on_s0(validator_config['validator-addr'], amount, node_config['endpoint']):
            log(f"{Typgpy.FAIL}Cannot create validator, {validator_config['validator-addr']} "
                f"does not have sufficient funds (need {amount} ONE). Checked {count} time(s).{Typgpy.ENDC}")
            if not _hard_reset_recovery:
                raise RuntimeError("Create Validator Error")
            log(f"{Typgpy.WARNING}Checking again in {check_interval} seconds.{Typgpy.ENDC}")
            time.sleep(check_interval)
        else:
            log(f"{Typgpy.OKGREEN}Address: {validator_config['validator-addr']} has enough funds{Typgpy.ENDC}")
            return


def _send_create_validator_tx():
    log(f"{Typgpy.OKBLUE}Sending create validator transaction...{Typgpy.ENDC}")
    passphrase = get_wallet_passphrase()
    count = 0
    while True:
        count += 1
        try:
            cmd = ['hmy', '--node', f'{node_config["endpoint"]}', 'staking', 'create-validator',
                   '--validator-addr', f'{validator_config["validator-addr"]}',
                   '--name', f'{validator_config["name"]}',
                   '--identity', f'{validator_config["identity"]}',
                   '--website', f'{validator_config["website"]}',
                   '--security-contact', f'{validator_config["security-contact"]}',
                   '--details', f'{validator_config["details"]}',
                   '--rate', f'{validator_config["rate"]}',
                   '--max-rate', f'{validator_config["max-rate"]}',
                   '--max-change-rate', f'{validator_config["max-change-rate"]}',
                   '--min-self-delegation', f'{validator_config["min-self-delegation"]}',
                   '--max-total-delegation', f'{validator_config["max-total-delegation"]}',
                   '--amount', f'{validator_config["amount"]}',
                   '--bls-pubkeys', f'{",".join(node_config["public-bls-keys"])}',
                   '--bls-pubkeys-dir', bls_key_dir, "--passphrase"]
            if validator_config["gas-price"]:
                cmd.extend(['--gas-price', f'{validator_config["gas-price"]}'])
            proc = cli.expect_call(cmd)
            pexpect_input_wallet_passphrase(proc, passphrase)
            proc.expect(pexpect.EOF)
            response = proc.before.decode()
            log(f"{Typgpy.OKBLUE}Create-validator transaction response: "
                f"{Typgpy.OKGREEN}{response}{Typgpy.ENDC}")
            return
        except (RuntimeError, TimeoutError, ConnectionError, subprocess.CalledProcessError) as e:
            log(f"{Typgpy.FAIL}Create-validator transaction failure (attempt {count}). Error: {e}{Typgpy.ENDC}")
            if not _hard_reset_recovery:
                raise e
            log(f"{Typgpy.WARNING}Trying again in {check_interval} seconds.{Typgpy.ENDC}")
            time.sleep(check_interval)


def _create_new_validator():
    _verify_prestaking_epoch()
    _verify_account_balance(Decimal(validator_config['amount']) + _balance_buffer)
    _send_create_validator_tx()


def _verify_node_sync():
    log(f"{Typgpy.OKBLUE}Verifying Node Sync...{Typgpy.ENDC}")
    wait_for_node_response("http://localhost:9500/", sleep=1, verbose=True)
    wait_for_node_response(node_config['endpoint'], sleep=1, verbose=True)
    curr_headers = blockchain.get_latest_headers()
    curr_epoch_shard = curr_headers['shard-chain-header']['epoch']
    curr_epoch_beacon = curr_headers['beacon-chain-header']['epoch']
    ref_epoch = blockchain.get_current_epoch(endpoint=node_config['endpoint'])
    has_looped = False
    if curr_epoch_shard < ref_epoch or curr_epoch_beacon < ref_epoch:
        prompt = "Waiting for node to sync. Deactivate validator? [Y]/n \n> "
        auto_interaction = 'Y' if _hard_reset_recovery else None
        if is_active_validator() and can_safe_stop_node() \
                and input_with_print(prompt, auto_interaction).lower() in {'y', 'yes'}:
            try:
                log(f"{Typgpy.OKBLUE}Deactivating validator until node is synced.{Typgpy.ENDC}")
                deactivate_validator()
            except (TimeoutError, ConnectionError, RuntimeError, subprocess.CalledProcessError) as e:
                log(f"{Typgpy.FAIL}Unable to deactivate validator {validator_config['validator-addr']}"
                    f"error {e}. Continuing...{Typgpy.ENDC}")
    while curr_epoch_shard < ref_epoch or curr_epoch_beacon < ref_epoch:
        sys.stdout.write(f"\rWaiting for node to sync: shard epoch ({curr_epoch_shard}/{ref_epoch}) "
                         f"& beacon epoch ({curr_epoch_beacon}/{ref_epoch})")
        sys.stdout.flush()
        has_looped = True
        time.sleep(check_interval)
        assert_no_invalid_blocks()
        curr_headers = blockchain.get_latest_headers()
        curr_epoch_shard = curr_headers['shard-chain-header']['epoch']
        curr_epoch_beacon = curr_headers['beacon-chain-header']['epoch']
        ref_epoch = blockchain.get_current_epoch(endpoint=node_config['endpoint'])
    if curr_epoch_shard > ref_epoch + 1 or curr_epoch_beacon > ref_epoch + 1:  # +1 for some slack on epoch change.
        log(f"{Typgpy.FAIL}Node epoch (shard: {curr_epoch_shard} beacon: {curr_epoch_beacon}) is greater than network "
            f"epoch ({ref_epoch}) which is not possible, is config correct?{Typgpy.ENDC}")
        if not _hard_reset_recovery:
            raise RuntimeError("Invalid node sync")
    if has_looped:
        log("")
    log(f"{Typgpy.OKGREEN}Node synced to current epoch...{Typgpy.ENDC}")
    try:
        prompt = "Node is synced and ready to validate. Activate validator? [Y]/n \n> "
        auto_interaction = 'Y' if _hard_reset_recovery else None
        if not is_active_validator() and input_with_print(prompt, auto_interaction).lower() in {'y', 'yes'}:
            activate_validator()
    except (TimeoutError, ConnectionError, RuntimeError, subprocess.CalledProcessError) as e:
        log(f"{Typgpy.FAIL}Unable to activate validator {validator_config['validator-addr']}"
            f"error {e}. Continuing...{Typgpy.ENDC}")


def _get_edit_validator_options():
    changeable_fields = {
        "details", "identity", "name", "security-contact", "website",
        "max-total-delegation", "min-self-delegation", "rate"
    }
    edit_validator_fields = {}
    for key, value in validator_config.items():
        if key in changeable_fields:
            edit_validator_fields[key] = validator_config[key]
    return edit_validator_fields


def _first_setup():
    """
    Initial setup done on first run of validator setup.
    """
    while True:
        try:
            setup_validator_config()
            assert_node_started(do_log=True)
            setup_wallet_passphrase()
            return
        except AssertionError as e:
            log(f"{Typgpy.WARNING}Assertion error: {e}{Typgpy.ENDC}")
            if input_with_print("Try again? [Y/n]\n> ").lower() not in {'y', 'yes'}:
                raise e


def get_validator_information():
    """
    Get the current validator information from the configured endpoint.
    """
    return staking.get_validator_information(validator_config['validator-addr'], endpoint=node_config['endpoint'])


def get_balances():
    """
    Get the balances of the configured validator (if possible)
    """
    balances = account.get_balance_on_all_shards(validator_config['validator-addr'], endpoint=node_config['endpoint'])
    for bal in balances:
        bal['balance'] = float(numbers.convert_atto_to_one(bal['balance']))
    return balances


def is_active_validator():
    """
    Default to false if exception to be defensive.
    """
    try:
        val_chain_info = get_validator_information()
        return val_chain_info['active-status'] == 'active'
    except (exceptions.RPCError, exceptions.RequestsError, exceptions.RequestsTimeoutError) as e:
        log(f"{Typgpy.WARNING}Could not fetch validator active status, error: {e}{Typgpy.ENDC}")
        return False


def deactivate_validator():
    try:
        all_val = staking.get_all_validator_addresses(endpoint=node_config['endpoint'])
        if validator_config["validator-addr"] in all_val:
            log(f"{Typgpy.OKBLUE}Deactivating validator{Typgpy.ENDC}")
            passphrase = get_wallet_passphrase()
            cmd = ['hmy', 'staking', 'edit-validator',
                   '--validator-addr', f'{validator_config["validator-addr"]}',
                   '--active', 'false', '--node', f'{node_config["endpoint"]}',
                   '--passphrase']
            if validator_config["gas-price"]:
                cmd.extend(['--gas-price', f'{validator_config["gas-price"]}'])
            proc = cli.expect_call(cmd)
            pexpect_input_wallet_passphrase(proc, passphrase)
            proc.expect(pexpect.EOF)
            response = proc.before.decode()
            log(f"{Typgpy.OKGREEN}Edit-validator response: {response}{Typgpy.ENDC}")
        else:
            log(f"{Typgpy.FAIL}Address {validator_config['validator-addr']} is not a validator!{Typgpy.ENDC}")
    except (TimeoutError, ConnectionError, RuntimeError, subprocess.CalledProcessError) as e:
        log(traceback.format_exc())
        log(f"{Typgpy.FAIL}{Typgpy.BOLD}Edit-validator error: {e}{Typgpy.ENDC}")
        if not _hard_reset_recovery:
            raise e
        log(f"{Typgpy.WARNING}{Typgpy.BOLD}Continuing...{Typgpy.ENDC}")


def activate_validator():
    try:
        all_val = staking.get_all_validator_addresses(endpoint=node_config['endpoint'])
        if validator_config["validator-addr"] in all_val:
            log(f"{Typgpy.OKBLUE}Activating validator{Typgpy.ENDC}")
            passphrase = get_wallet_passphrase()
            cmd = ['hmy', 'staking', 'edit-validator', '--validator-addr', f'{validator_config["validator-addr"]}',
                   '--active', 'true', '--node', f'{node_config["endpoint"]}',
                   '--passphrase']
            if validator_config["gas-price"]:
                cmd.extend(['--gas-price', f'{validator_config["gas-price"]}'])
            proc = cli.expect_call(cmd)
            pexpect_input_wallet_passphrase(proc, passphrase)
            proc.expect(pexpect.EOF)
            response = proc.before.decode()
            log(f"{Typgpy.OKGREEN}Edit-validator response: {response}{Typgpy.ENDC}")
        else:
            log(f"{Typgpy.FAIL}Address {validator_config['validator-addr']} is not a validator!{Typgpy.ENDC}")
    except (TimeoutError, ConnectionError, RuntimeError, subprocess.CalledProcessError) as e:
        log(traceback.format_exc())
        log(f"{Typgpy.FAIL}{Typgpy.BOLD}Edit-validator error: {e}{Typgpy.ENDC}")
        if not _hard_reset_recovery:
            raise e
        log(f"{Typgpy.WARNING}{Typgpy.BOLD}Continuing...{Typgpy.ENDC}")


def collect_reward():
    try:
        all_val = staking.get_all_validator_addresses(endpoint=node_config['endpoint'])
        if validator_config["validator-addr"] in all_val:
            log(f"{Typgpy.OKBLUE}Collecting rewards{Typgpy.ENDC}")
            passphrase = get_wallet_passphrase()
            cmd = ['hmy', 'staking', 'collect-rewards', '--delegator-addr', validator_config['validator-addr'],
                   '--node', f'{node_config["endpoint"]}', '--passphrase']
            if validator_config["gas-price"]:
                cmd.extend(['--gas-price', f'{validator_config["gas-price"]}'])
            proc = cli.expect_call(cmd)
            pexpect_input_wallet_passphrase(proc, passphrase)
            proc.expect(pexpect.EOF)
            response = proc.before.decode()
            log(f"{Typgpy.OKGREEN}Collect rewards response: {response}{Typgpy.ENDC}")
        else:
            log(f"{Typgpy.FAIL}Address {validator_config['validator-addr']} is not a validator!{Typgpy.ENDC}")
    except (TimeoutError, ConnectionError, RuntimeError, subprocess.CalledProcessError) as e:
        log(traceback.format_exc())
        log(f"{Typgpy.FAIL}{Typgpy.BOLD}Edit-validator error: {e}{Typgpy.ENDC}")
        if not _hard_reset_recovery:
            raise e
        log(f"{Typgpy.WARNING}{Typgpy.BOLD}Continuing...{Typgpy.ENDC}")


def check_and_activate():
    """
    Return True when attempted to activate, otherwise return False.
    """
    try:
        if not is_active_validator():
            log(f"{Typgpy.FAIL}Node not active, reactivating...{Typgpy.ENDC}")
            curr_headers = blockchain.get_latest_headers()
            curr_epoch_shard = curr_headers['shard-chain-header']['epoch']
            curr_epoch_beacon = curr_headers['beacon-chain-header']['epoch']
            wait_for_node_response(node_config['endpoint'], tries=900, sleep=1, verbose=False)  # Try for 15 min
            ref_epoch = blockchain.get_current_epoch(endpoint=node_config['endpoint'])
            if curr_epoch_shard == ref_epoch and curr_epoch_beacon == ref_epoch:
                activate_validator()
                return True
            else:
                log(f"{Typgpy.WARNING}Node not synced, did NOT activate node.{Typgpy.ENDC}")
                return False
    except (TimeoutError, ConnectionError, RuntimeError, subprocess.CalledProcessError) as e:
        log(traceback.format_exc())
        log(f"{Typgpy.FAIL}Unable to activate validator {validator_config['validator-addr']}"
            f"error {e}. Continuing...{Typgpy.ENDC}")
        if not _hard_reset_recovery:  # Do not throw error on hard resets.
            raise e
    return False


def setup(hard_reset_recovery=False):
    log(f"{Typgpy.HEADER}Starting validator setup...{Typgpy.ENDC}")
    old_logging_handlers = _interaction_preprocessor(hard_reset_recovery)
    log(f"{Typgpy.OKBLUE}Using BLS key(s): {Typgpy.OKGREEN}{node_config['public-bls-keys']}{Typgpy.ENDC}")
    try:
        if not hard_reset_recovery:
            _first_setup()
        wait_for_node_response(node_config['endpoint'], verbose=True, tries=120, sleep=1)  # Try for 2 min
        all_val_address = staking.get_all_validator_addresses(endpoint=node_config['endpoint'])
        if validator_config['validator-addr'] in all_val_address:
            log(f"{Typgpy.WARNING}{validator_config['validator-addr']} already in list of validators!{Typgpy.ENDC}")
            validator_info = get_validator_information()
            keys_on_chain = set(validator_info['validator']['bls-public-keys'])
            if all(k in keys_on_chain for k in node_config["public-bls-keys"]):
                log(f"{Typgpy.OKBLUE}{Typgpy.BOLD}No BLS key(s) to add to validator!{Typgpy.ENDC}")
            else:
                prompt = "Add BLS key(s) to existing validator? [Y]/n \n> "
                if input_with_print(prompt, 'Y' if hard_reset_recovery else None).lower() in {'y', 'yes'}:
                    log(f"{Typgpy.HEADER}{Typgpy.BOLD}Editing validator...{Typgpy.ENDC}")
                    _add_bls_key_to_validator()
        elif validator_config['validator-addr'] not in all_val_address:
            prompt = "Create validator? [Y]/n \n> "
            if input_with_print(prompt, 'Y' if hard_reset_recovery else None).lower() in {'y', 'yes'}:
                log(f"{Typgpy.HEADER}{Typgpy.BOLD}Creating new validator...{Typgpy.ENDC}")
                _create_new_validator()
        else:
            node_config['no-validator'] = True
        log(f"{Typgpy.HEADER}{Typgpy.BOLD}Finished setting up validator!{Typgpy.ENDC}")
        _verify_node_sync()
        logging.getLogger('AutoNode').handlers = old_logging_handlers
    except Exception as e:
        log(traceback.format_exc())
        logging.getLogger('AutoNode').handlers = old_logging_handlers
        if not _hard_reset_recovery:
            raise e
        else:
            log(f"{Typgpy.FAIL}{Typgpy.BOLD}Validator creation error: {e}{Typgpy.ENDC}")
            log(f"{Typgpy.WARNING}{Typgpy.BOLD}Continuing...{Typgpy.ENDC}")


def update_info(hard_reset_recovery=False):
    old_logging_handlers = _interaction_preprocessor(hard_reset_recovery)
    address = validator_config['validator-addr']
    try:
        all_val_address = staking.get_all_validator_addresses(endpoint=node_config['endpoint'])
        if address not in all_val_address:
            log(f"{Typgpy.WARNING}Cannot edit validator information, validator "
                f"{Typgpy.OKGREEN}{address}{Typgpy.WARNING} is not a validator!{Typgpy.ENDC}")
            if hard_reset_recovery:
                return  # clean exit for hard resets
            raise RuntimeError("Validator does not exist")
        fields = _get_edit_validator_options()
        if fields:
            log(f"{Typgpy.OKBLUE}Updating validator information for {address}: "
                f"{Typgpy.OKGREEN}{json.dumps(fields, indent=2)}{Typgpy.ENDC}")
            passphrase = get_wallet_passphrase()
            cmd = ['hmy', '--node', f'{node_config["endpoint"]}', 'staking', 'edit-validator',
                   '--validator-addr', f'{address}', '--passphrase']
            for key, value in fields.items():
                cmd.extend([f'--{key}', f'{value}'])
            proc = cli.expect_call(cmd)
            pexpect_input_wallet_passphrase(proc, passphrase)
            proc.expect(pexpect.EOF)
            response = proc.before.decode()
            log(f"{Typgpy.OKBLUE}Edit-validator transaction response: {Typgpy.OKGREEN}{response}{Typgpy.ENDC}")
        logging.getLogger('AutoNode').handlers = old_logging_handlers
    except Exception as e:
        log(traceback.format_exc())
        logging.getLogger('AutoNode').handlers = old_logging_handlers
        if not _hard_reset_recovery:
            raise e
        else:
            log(f"{Typgpy.FAIL}{Typgpy.BOLD}Edit-validator error: {e}{Typgpy.ENDC}")
            log(f"{Typgpy.WARNING}{Typgpy.BOLD}Continuing...{Typgpy.ENDC}")


def remove_bls_key(key):
    """
    Remove the BLS key (if present) from the validator
    """
    assert isinstance(key, str)
    key = key.replace("0x", "").strip()
    keys_on_chain = [x.replace("0x", "").strip() for x in get_validator_information()['validator']['bls-public-keys']]
    if key not in keys_on_chain:
        log(f"{Typgpy.WARNING}BLS key {key} is not present on validator!")
        return
    passphrase = get_wallet_passphrase()
    log(f"{Typgpy.OKBLUE}Removing BLS key {Typgpy.OKGREEN}{key}{Typgpy.ENDC}")
    proc = cli.expect_call(['hmy', '--node', node_config['endpoint'], 'staking', 'edit-validator',
                            '--validator-addr', validator_config['validator-addr'],
                            '--remove-bls-key', key, '--passphrase'])
    pexpect_input_wallet_passphrase(proc, passphrase)
    proc.expect(pexpect.EOF)
    response = proc.before.decode()
    log(f"{Typgpy.OKGREEN}Edit-validator transaction response: {response}{Typgpy.ENDC}")


def can_safe_stop_node(safe=False):
    """
    Determine if a node can be stopped. Conditions:
    If elected, BLS keys must not be earning (if present), otherwise can shutdown.

    If `safe`, return false on exception, otherwise raise any exception
    """
    try:
        if node_config['no-validator']:
            return True
        addr = validator_config['validator-addr']
        all_val_address = staking.get_all_validator_addresses(endpoint=node_config['endpoint'])
        if addr in all_val_address and not get_validator_information()['currently-in-committee']:
            return True
        return not is_signing()
    except Exception as e:
        if safe:
            return False
        raise e
